import hashlib
import logging
import uuid
from textwrap import dedent

from django.conf import settings
from django.db import transaction
from django.db.models import Exists, OuterRef
from django.shortcuts import reverse
from django.utils import timezone
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.ponos.keys import check_agent_key
from arkindex.ponos.models import (
    FINAL_STATES,
    GPU,
    ActionType,
    Agent,
    Artifact,
    Farm,
    Secret,
    State,
    Task,
    task_token_default,
)
from arkindex.ponos.serializer_fields import Base64Field, CurrentProcessDefault, PublicKeyField
from arkindex.ponos.signals import task_failure
from arkindex.project.serializer_fields import EnumField
from arkindex.project.triggers import notify_process_completion

logger = logging.getLogger(__name__)


class FarmSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Farm` instance.
    """

    class Meta:
        model = Farm
        fields = (
            "id",
            "name",
        )


class GPUSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.GPU` instance for public access.
    """

    class Meta:
        model = GPU
        fields = (
            "id",
            "index",
            "name",
            "ram_total",
        )
        read_only_fields = (
            "id",
            "index",
            "name",
            "ram_total",
        )


class AgentLightSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Agent` instance for public access.
    """

    farm = FarmSerializer()
    gpus = GPUSerializer(many=True, read_only=True)

    class Meta:
        model = Agent
        fields = (
            "id",
            "farm",
            "hostname",
            "cpu_cores",
            "cpu_frequency",
            "ram_total",
            "cpu_load",
            "ram_load",
            "last_ping",
            "gpus",
        )
        read_only_fields = ("id", "last_ping")


class TaskLightSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance without logs or agent information.
    Used to list tasks inside a process.
    """
    state = EnumField(
        State,
        help_text=dedent("""
            Allowed transitions for the state of a task by an agent are defined below:

                Pending ⟶ Running ⟶ Completed
                  └⟶ Error   ├⟶ Failed
                             └⟶ Error
                Stopping ⟶ Stopped
                  └⟶ Error
        """).strip(),
    )

    class Meta:
        model = Task
        fields = (
            "id",
            "run",
            "depth",
            "slug",
            "state",
            "parents",
            "shm_size",
        )
        read_only_fields = (
            "id",
            "run",
            "depth",
            "slug",
            "parents",
            "shm_size",
        )

    def validate_state(self, state):
        allowed_transitions = {
            State.Unscheduled: [State.Pending],
            State.Pending: [State.Running, State.Error],
            State.Running: [State.Completed, State.Failed, State.Error],
            State.Stopping: [State.Stopped, State.Error],
        }
        if self.instance and state not in allowed_transitions.get(self.instance.state, []):
            raise ValidationError(f"Transition from state {self.instance.state} to state {state} is forbidden.")
        return state


class TaskSerializer(TaskLightSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance with logs and agent information.
    """

    logs = serializers.SerializerMethodField()
    full_log = serializers.SerializerMethodField()
    agent = AgentLightSerializer(read_only=True, allow_null=True)
    gpu = GPUSerializer(read_only=True, allow_null=True)
    extra_files = serializers.DictField(read_only=True)

    class Meta(TaskLightSerializer.Meta):
        fields = TaskLightSerializer.Meta.fields + (
            "logs",
            "full_log",
            "agent",
            "gpu",
            "extra_files",
        )
        read_only_fields = TaskLightSerializer.Meta.read_only_fields + (
            "logs",
            "full_log",
            "agent",
            "gpu",
            "extra_files",
        )

    @extend_schema_field(serializers.CharField())
    def get_logs(self, task) -> str:
        return task.logs.latest

    @extend_schema_field(serializers.URLField())
    def get_full_log(self, task) -> str:
        return task.logs.s3_url

    def update(self, instance: Task, validated_data) -> Task:
        """
        Perform updates on a task instance.
        If a task gets marked as :attr:`~arkindex.ponos.models.State.Completed`, this will set
        its child tasks to :attr:`~arkindex.ponos.models.State.Pending` if all of their dependencies are met.
        """
        instance = super().update(instance, validated_data)

        if instance.state not in FINAL_STATES:
            return instance

        with transaction.atomic(using="default"):
            # When a task is completed, set its children to be started next
            if instance.state == State.Completed:
                (
                    instance.children
                    # We need to use the default Database here to avoid a stale read
                    # as the parent instance has been updated just before
                    .using("default")
                    .filter(state=State.Unscheduled)
                    # Do not update any child whose dependencies include some tasks that are not completed
                    .exclude(Exists(Task.objects.filter(children=OuterRef("pk")).exclude(state=State.Completed)))
                    .update(state=State.Pending)
                )

            # When some tasks have failed, tasks that depend on them will remain unscheduled.
            # When a task is completed, its children will be set to pending.
            # All children have a depth above zero, since they have at least one dependency.
            # This means a process can be considered finished when no task is pending (about to run) or running,
            # except for tasks with a depth of zero, which can be unscheduled until they get assigned to an agent.
            if not (
                instance.process.tasks
                .using("default")
                .filter(run=instance.run)
                .exclude(state__in=FINAL_STATES)
                .exclude(state=State.Unscheduled, depth__gt=0)
                .exists()
            ):
                instance.process.finished = timezone.now()
                instance.process.save()

                notify_process_completion(instance.process)

        # We already checked earlier that the task was in a final state.
        # If this state is both final and not completed, then we should trigger the task failure signal.
        if instance.state != State.Completed:
            task_failure.send_robust(self.__class__, task=instance)

        return instance


class TaskTinySerializer(TaskSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance with only its state.
    Used by humans to update a task.
    """

    state = EnumField(
        State,
        help_text=dedent("""
            Allowed transitions for the state of a task by a user are defined below:

                Completed ⟶ Pending
                Failed ⟶ Pending
                Error ⟶ Pending
                Stopped ⟶ Pending
                Running ⟶ Stopping
        """).strip(),
    )

    class Meta:
        model = Task
        fields = ("id", "state")
        read_only_fields = ("id",)

    def validate_state(self, state):
        """
        Only allow a user to manually stop or retry a task
        """
        allowed_transitions = {
            state: [State.Pending] for state in FINAL_STATES
        }
        allowed_transitions.update({State.Running: [State.Stopping]})
        if self.instance and state not in allowed_transitions.get(self.instance.state, []):
            raise ValidationError(f"Transition from state {self.instance.state} to state {state} is forbidden.")
        return state

    def update(self, instance: Task, validated_data) -> Task:
        new_state = validated_data.get("state")

        if new_state == State.Pending:
            if instance.state == State.Unscheduled:
                # Prevent a user from restarting a task that was never assigned to an agent.
                raise ValidationError({
                    "state": [f"Transition from state {State.Unscheduled} to state {State.Pending} is forbidden."]
                })
            # Restart the task
            instance.agent = None
            instance.gpu = None
            # Un-finish the process since a task will run again
            instance.process.finished = None
            instance.process.save()

        if new_state in FINAL_STATES and new_state != State.Completed:
            task_failure.send_robust(self.__class__, task=instance)

        return super().update(instance, validated_data)


class AgentStateSerializer(AgentLightSerializer):
    """
    Serialize an :class:`~arkindex.ponos.models.Agent` with its state information
    And the GPU state
    """

    running_tasks_count = serializers.IntegerField(min_value=0)

    class Meta(AgentLightSerializer.Meta):
        fields = AgentLightSerializer.Meta.fields + ("active", "running_tasks_count")


class AgentDetailsSerializer(AgentLightSerializer):
    """
    Serialize an :class:`~arkindex.ponos.models.Agent` with its running tasks
    """

    running_tasks = serializers.SerializerMethodField()

    class Meta(AgentLightSerializer.Meta):
        fields = AgentLightSerializer.Meta.fields + ("active", "running_tasks")

    @extend_schema_field(TaskLightSerializer(many=True))
    def get_running_tasks(self, agent):
        running_tasks = agent.tasks.filter(state=State.Running)
        task_serializer = TaskLightSerializer(
            instance=running_tasks, many=True, context=self.context
        )
        return task_serializer.data


class GPUCreateSerializer(serializers.Serializer):
    """
    Serialize a :class:`~arkindex.ponos.models.GPU` attached to an Agent
    """

    id = serializers.UUIDField()
    name = serializers.CharField()
    index = serializers.IntegerField()
    ram_total = serializers.IntegerField()


class AgentCreateSerializer(serializers.ModelSerializer):
    """
    Serializer used to register a new :class:`~arkindex.ponos.models.Agent`.
    """

    public_key = PublicKeyField(write_only=True)
    derivation = Base64Field(write_only=True)
    access_token = serializers.CharField(read_only=True, source="token.access_token")
    refresh_token = serializers.CharField(read_only=True, source="token")
    gpus = GPUCreateSerializer(many=True)

    class Meta:
        model = Agent
        fields = (
            "id",
            "farm",
            "access_token",
            "refresh_token",
            "public_key",
            "derivation",
            "hostname",
            "cpu_cores",
            "cpu_frequency",
            "ram_total",
            "cpu_load",
            "ram_load",
            "last_ping",
            "gpus",
        )
        read_only_fields = (
            "id",
            "access_token",
            "refresh_token",
            "last_ping",
        )

    def create(self, validated_data):
        """Create the required agent and its GPUs"""
        gpus = validated_data.pop("gpus", None)

        with transaction.atomic(using="default"):

            # Create Agent as usual
            agent = super().create(validated_data)

            # Create or update GPUs
            # When an agent's private key is regenerated on the same host, an existing GPU UUID might be sent,
            # so we need to handle existing GPUs to avoid errors.
            for gpu_data in gpus:
                GPU.objects.update_or_create(
                    id=gpu_data.pop("id"), defaults=dict(agent=agent, **gpu_data)
                )

        return agent

    def update(self, instance, validated_data):
        """Create the required agent and its GPUs"""
        gpus = validated_data.pop("gpus", None)

        with transaction.atomic(using="default"):

            # Update Agent as usual
            agent = super().update(instance, validated_data)

            # Delete existing GPUs
            agent.gpus.all().delete()

            # Create Gpus - the ID should not evolve as it's provided
            # by the host
            for gpu_data in gpus:
                agent.gpus.create(**gpu_data)

        return agent

    def validate(self, data):
        if not check_agent_key(
            data["public_key"], data["derivation"], data["farm"].seed
        ):
            raise serializers.ValidationError("Key verification failed")
        del data["derivation"]

        # Turn the public key back into a string to be saved in the agent
        # TODO: Use a custom Django model field for this?
        data["public_key"] = PublicKeyField().to_representation(data["public_key"])

        # Generate the agent ID as a MD5 hash of its public key
        data["id"] = uuid.UUID(
            hashlib.md5(data["public_key"].encode("utf-8")).hexdigest()
        )

        # Set last ping as agent is about to be registered
        data["last_ping"] = timezone.now()

        return data


class ActionSerializer(serializers.Serializer):
    """
    Serializes an :const:`~arkindex.ponos.models.Action` instance.
    """

    action = EnumField(ActionType)
    task_id = serializers.SerializerMethodField()

    @extend_schema_field(serializers.UUIDField(allow_null=True))
    def get_task_id(self, obj):
        if not obj.task:
            return
        return str(obj.task.id)


class AgentActionsSerializer(serializers.Serializer):
    """
    Serializes multiple next actions for an agent.
    """

    actions = ActionSerializer(many=True, source="next_actions")


class TaskDefinitionSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance
    to provide startup information to assigned agents.
    """

    env = serializers.DictField(default={})
    process_id = serializers.UUIDField()
    agent_id = serializers.PrimaryKeyRelatedField(queryset=Agent.objects.all())
    image_artifact_url = serializers.SerializerMethodField()
    s3_logs_put_url = serializers.SerializerMethodField()
    extra_files = serializers.DictField(default={})
    state = EnumField(State)

    @extend_schema_field(serializers.URLField(allow_null=True))
    def get_image_artifact_url(self, task):
        """Build url on the API to get a fresh download link"""
        if not task.image_artifact:
            return
        return self.context["request"].build_absolute_uri(
            reverse(
                "api:task-artifact-download",
                kwargs={
                    "pk": task.image_artifact.task_id,
                    "path": task.image_artifact.path,
                },
            )
        )

    @extend_schema_field(serializers.URLField(allow_null=True))
    def get_s3_logs_put_url(self, obj):
        if "request" not in self.context or self.context["request"].user != obj.agent:
            return
        return obj.logs.s3_put_url

    class Meta:
        model = Task
        fields = (
            "id",
            "slug",
            "image",
            "command",
            "env",
            "shm_size",
            "has_docker_socket",
            "image_artifact_url",
            "agent_id",
            "s3_logs_put_url",
            "parents",
            "process_id",
            "gpu_id",
            "extra_files",
            "state",
        )
        read_only_fields = fields


class ArtifactSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Artifact` instance to allow
    agents to create them, and users to list and retrieve them
    """

    s3_put_url = serializers.SerializerMethodField()

    class Meta:
        model = Artifact
        fields = (
            "id",
            "path",
            "size",
            "content_type",
            "s3_put_url",
            "created",
            "updated",
        )
        read_only_fields = (
            "id",
            "s3_put_url",
            "created",
            "updated",
        )

    @extend_schema_field(serializers.URLField(allow_null=True))
    def get_s3_put_url(self, obj):
        """Only add the PUT url to store the file for agents during creation"""
        request = self.context.get("request")
        if not request or request.method != "POST":
            return
        return obj.s3_put_url

    def validate_path(self, path):
        """Check that no artifacts with this path already exist in DB"""
        task = self.context["view"].task
        if task.artifacts.filter(path=path).exists():
            raise ValidationError("An artifact with this path already exists")

        return path


class NewTaskSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance to permit creation by a parent.
    """

    process = serializers.HiddenField(default=CurrentProcessDefault())
    process_id = serializers.UUIDField(read_only=True)
    command = serializers.CharField(required=False)
    env = serializers.DictField(
        child=serializers.CharField(), required=False, default={}
    )
    has_docker_socket = serializers.BooleanField(default=False)

    class Meta:
        model = Task
        fields = (
            "id",
            "process",
            "process_id",
            "slug",
            "parents",
            "image",
            "command",
            "env",
            "run",
            "depth",
            "has_docker_socket",
        )
        read_only_fields = (
            "id",
            "run",
            "depth",
            "process_id",
        )

    def validate(self, data):
        parents = data["parents"]

        ids = {parent.process_id for parent in parents}
        if len(ids) != 1 or str(ids.pop()) != str(data["process"].id):
            raise ValidationError(
                "All parents must be in the same process as the child task"
            )

        runs = {parent.run for parent in parents}
        if len(runs) != 1:
            raise ValidationError(
                "All parents must have the same run in the given process"
            )
        data["run"] = runs.pop()

        if Task.objects.filter(
            process=data["process"], run=data["run"], slug=data["slug"]
        ).exists():
            raise ValidationError(
                f"A task with the `{data['slug']}` slug already exists in run {data['run']}."
            )

        data["depth"] = max(parent.depth for parent in parents) + 1

        # Build task environment from PONOS_DEFAULT_ENV if the env is not (fully) defined in data
        data["env"] = {**settings.PONOS_DEFAULT_ENV, **data["env"]}
        data["env"]["ARKINDEX_PROCESS_ID"] = str(data["process"].id)
        # Get corpus id from any of the parent tasks since it's the same for all of them
        # TODO: Get the corpus ID from the process using a ForeignKeyField on the serializer and data["process"]
        data["env"]["ARKINDEX_CORPUS_ID"] = parents[0].env.get("ARKINDEX_CORPUS_ID", None)

        # Set the task token manually so that we can immediately copy it to the environment variables,
        # just like what is done in Workflow.build_tasks()
        data["token"] = task_token_default()
        data["env"]["ARKINDEX_TASK_TOKEN"] = data["token"]

        return super().validate(data)


class ClearTextSecretSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Secret` instance with its content in cleartext.
    """

    content = serializers.SerializerMethodField(read_only=True)

    def get_content(self, secret) -> str:
        return secret.decrypt()

    class Meta:
        model = Secret
        fields = ("id", "name", "content")
        read_only_fields = ("id", "content")
