from collections import defaultdict

from django.db.models import Exists, OuterRef
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied, ValidationError

from arkindex.documents.models import Corpus
from arkindex.process.models import Process, ProcessDataset, ProcessMode, Task
from arkindex.project.mixins import ProcessACLMixin
from arkindex.training.models import Dataset
from arkindex.training.serializers import DatasetSerializer
from arkindex.users.models import Role


def _dataset_id_from_context(serializer_field):
    return serializer_field.context.get("view").kwargs["dataset"]


def _process_id_from_context(serializer_field):
    return serializer_field.context.get("view").kwargs["process"]


_dataset_id_from_context.requires_context = True
_process_id_from_context.requires_context = True


class ProcessDatasetSerializer(ProcessACLMixin, serializers.ModelSerializer):
    process_id = serializers.HiddenField(
        write_only=True,
        default=_process_id_from_context
    )
    dataset_id = serializers.HiddenField(
        write_only=True,
        default=_dataset_id_from_context
    )
    dataset = DatasetSerializer(read_only=True)
    sets = serializers.ListField(
        child=serializers.CharField(max_length=50),
        required=False,
        allow_null=False,
        allow_empty=False,
        min_length=1
    )

    class Meta:
        model = ProcessDataset
        fields = ("dataset_id", "dataset", "process_id", "id", "sets", )
        read_only_fields = ("process_id", "id", )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get("request"):
            # Do not raise Error in order to create OpenAPI schema
            return
        # Required for the ProcessACLMixin and readable corpora
        self._user = self.context["request"].user

    def validate(self, data):
        request_method = self.context["request"].method
        errors = defaultdict(list)

        # Validate process
        process_qs = (
            Process
            .objects
            # Avoid a stale read when adding a dataset right after creating a process
            .using("default")
            # Required for ACL checks
            .select_related("corpus", "revision__repo")
            # Required to check for a process that has already started
            .annotate(has_tasks=Exists(Task.objects.filter(process=OuterRef("pk"))))
        )
        if not self.instance:
            try:
                process = process_qs.get(pk=data["process_id"])
            except Process.DoesNotExist:
                raise ValidationError({"process": [f'Invalid pk "{str(data["process_id"])}" - object does not exist.']})
        else:
            process = self.instance.process
        access = self.process_access_level(process)
        if not access or not (access >= Role.Admin.value):
            raise PermissionDenied(detail="You do not have admin access to this process.")

        # Validate dataset
        if not self.instance:
            if request_method == "DELETE":
                # Allow deleting ProcessDatasets even if the user looses access to the corpus
                dataset_qs = Dataset.objects.all()
            else:
                dataset_qs = Dataset.objects.filter(corpus__in=Corpus.objects.readable(self._user))
            try:
                dataset = dataset_qs.select_related("creator").get(pk=data["dataset_id"])
            except Dataset.DoesNotExist:
                raise ValidationError({"dataset": [f'Invalid pk "{str(data["dataset_id"])}" - object does not exist.']})
        else:
            dataset = self.instance.dataset
        data["dataset"] = dataset

        if process.mode != ProcessMode.Dataset:
            errors["process"].append('Datasets can only be added to or removed from processes of mode "dataset".')

        if process.has_tasks:
            errors["process"].append("Datasets cannot be updated on processes that have already started.")

        if self.context["request"].method == "POST" and process.datasets.filter(id=dataset.id).exists():
            errors["dataset"].append("This dataset is already selected in this process.")

        # Validate sets
        sets = data.get("sets")
        if not sets or len(sets) == 0:
            if not self.instance:
                data["sets"] = dataset.sets
            else:
                errors["sets"].append("This field cannot be empty.")
        else:
            if any(s not in dataset.sets for s in sets):
                errors["sets"].append("The specified sets must all exist in the specified dataset.")
            if len(set(sets)) != len(sets):
                errors["sets"].append("Sets must be unique.")

        if errors:
            raise ValidationError(errors)

        return data
