import hashlib
import uuid
from datetime import datetime, timezone
from unittest.mock import patch

from django.urls import reverse
from rest_framework import status
from rest_framework.exceptions import ValidationError

from arkindex.ponos.authentication import AgentUser
from arkindex.ponos.models import Farm, State
from arkindex.process.models import (
    ActivityState,
    ProcessMode,
    Repository,
    Worker,
    WorkerActivityState,
    WorkerRun,
    WorkerType,
    WorkerVersion,
)
from arkindex.process.signals import _list_ancestors
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_public_key
from arkindex.training.models import Model


class TestSignals(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = Repository.objects.get(url="http://my_repo.fake/workers/worker")
        cls.rev_1 = cls.repo.revisions.get()
        cls.rev_2 = cls.repo.revisions.create(
            hash="2",
            message="commit message",
            author="bob",
        )

        cls.worker_type_1 = WorkerType.objects.get(slug="recognizer")

        cls.worker_1 = Worker.objects.create(
            repository=cls.repo,
            name="Worker 1",
            slug="worker_1",
            type=cls.worker_type_1
        )
        cls.version_1 = WorkerVersion.objects.create(
            worker=cls.worker_1,
            revision=cls.rev_1,
            configuration={"test": "test1"}
        )
        cls.version_2 = WorkerVersion.objects.create(
            worker=cls.worker_1,
            revision=cls.rev_2,
            configuration={"test": "test2"}
        )

        cls.farm = Farm.objects.first()
        cls.process_1 = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            farm=cls.farm,
        )
        cls.run_1 = cls.process_1.worker_runs.create(
            version=cls.version_1,
            parents=[],
        )
        cls.process_2 = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            farm=cls.farm,
        )
        pubkey = build_public_key()
        cls.agent = AgentUser.objects.create(
            id=uuid.UUID(hashlib.md5(pubkey.encode("utf-8")).hexdigest()),
            farm=cls.farm,
            hostname="ghostname",
            cpu_cores=2,
            cpu_frequency=1e9,
            public_key=pubkey,
            ram_total=2e9,
            last_ping=datetime.now(),
            cpu_load=.1,
            ram_load=.1e9,
        )

    def test_worker_run_check_parents_recursive(self):
        run_2 = self.process_1.worker_runs.create(
            version=self.version_2,
            parents=[],
        )
        self.assertListEqual(run_2.parents, [])

        run_2.parents = [str(run_2.id)]
        with self.assertRaises(ValidationError):
            run_2.save()

        run_2.refresh_from_db()
        self.assertListEqual(run_2.parents, [])

    def test_worker_run_check_parents_distinct_process(self):
        self.assertEqual(len(self.process_1.worker_runs.all()), 1)
        with self.assertRaises(ValidationError):
            self.process_2.worker_runs.create(
                version=self.version_2,
                parents=[str(self.run_1.id)],
            )
        self.process_1.refresh_from_db()

        self.assertEqual(len(self.process_1.worker_runs.all()), 1)

    def test_worker_run_check_parents_inexsistant_uuid(self):
        self.assertEqual(len(self.process_1.worker_runs.all()), 1)
        with self.assertRaises(ValidationError):
            self.process_2.worker_runs.create(
                version=self.version_2,
                parents=["12341234-1234-1234-1234-123412341234"],
            )
        self.process_1.refresh_from_db()

        self.assertEqual(len(self.process_1.worker_runs.all()), 1)

    def test_list_ancestors_no_parents(self):
        ancestors = _list_ancestors(WorkerRun.objects.none(), {})

        self.assertListEqual(ancestors, [])

    def test_list_ancestors_one_level(self):
        # --- TREE ---
        #
        #   B  C  D
        #    \ | /
        #     \|/
        #      A
        ancestors = _list_ancestors({"B": [], "C": [], "D": []}, ["B", "C", "D"])

        self.assertListEqual(ancestors, ["B", "C", "D"])

    def test_list_ancestors_three_level(self):
        # --- TREE ---
        #
        #   H     I  J
        #    \    | /
        #     \   |/
        #   E  F  G
        #    \ |  |
        #     \|  |
        #   B  C  D
        #    \ | /
        #     \|/
        #      A
        ancestors = _list_ancestors(
            {"B": [], "C": ["E", "F"], "D": ["G"], "E": [], "F": ["H"], "G": ["I", "J"], "H": [], "I": [], "J": []},
            ["B", "C", "D"]
        )

        self.assertListEqual(ancestors, ["B", "C", "E", "F", "H", "D", "G", "I", "J"])

    def test_worker_run_small_cycle(self):
        # --- TREE (unallowed cycle) ---
        #
        #             run_1
        #             |   |
        #             |   |
        #             run_2

        run_2 = self.process_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )

        self.run_1.parents = [run_2.id]
        with self.assertRaises(ValidationError):
            self.run_1.save()
        self.run_1.refresh_from_db()

        self.assertListEqual(self.run_1.parents, [])

    def test_worker_run_big_cycle(self):
        # --- TREE (unallowed cycle) ---
        #
        #             run_1
        #           /       \
        #        run_5     run_2
        #          |         |
        #           \       /
        #         run_4---run_3

        rev_3 = self.repo.revisions.create(
            hash="3",
            message="commit message",
            author="dylan",
        )
        rev_4 = self.repo.revisions.create(
            hash="4",
            message="commit message",
            author="pink",
        )
        rev_5 = self.repo.revisions.create(
            hash="5",
            message="commit message",
            author="floyd",
        )
        version_3 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_3,
            configuration={"test": "test3"}
        )
        version_4 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_4,
            configuration={"test": "test4"}
        )
        version_5 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_5,
            configuration={"test": "test5"}
        )

        run_2 = self.process_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )
        run_3 = self.process_1.worker_runs.create(
            version=version_3,
            parents=[run_2.id],
        )
        run_4 = self.process_1.worker_runs.create(
            version=version_4,
            parents=[run_3.id],
        )
        run_5 = self.process_1.worker_runs.create(
            version=version_5,
            parents=[run_4.id],
        )

        self.run_1.parents = [run_5.id]
        with self.assertNumQueries(2), self.assertRaises(ValidationError):
            self.run_1.save()
        self.run_1.refresh_from_db()

        self.assertListEqual(self.run_1.parents, [])

    def test_worker_run_no_cycle(self):
        rev_3 = self.repo.revisions.create(
            hash="3",
            message="commit message",
            author="dylan",
        )
        version_3 = WorkerVersion.objects.create(
            worker=self.worker_1,
            revision=rev_3,
            configuration={"test": "test3"}
        )

        run_2 = self.process_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )
        run_3 = self.process_1.worker_runs.create(
            version=version_3,
            parents=[run_2.id],
        )

        run_3.parents.append(self.run_1.id)
        with self.assertNumQueries(4):
            run_3.save()
        run_3.refresh_from_db()

        self.assertListEqual(run_3.parents, [run_2.id, self.run_1.id])

    def test_worker_run_check_parents(self):
        self.assertEqual(len(self.process_1.worker_runs.all()), 1)
        run_2 = self.process_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )

        self.assertEqual(len(self.process_1.worker_runs.all()), 2)
        self.assertListEqual(list(self.process_1.worker_runs.all()), [
            self.run_1,
            run_2,
        ])

    def test_worker_run_generate_summary(self):
        run = self.process_1.worker_runs.create(
            version=self.version_2,
            parents=[self.run_1.id],
        )

        self.assertIsNotNone(run.summary)

    @patch("arkindex.ponos.serializers.TaskSerializer.get_logs")
    @patch("arkindex.ponos.tasks.notify_process_completion.delay")
    def test_task_failure_updates_activities(self, notify_mock, get_logs_mock):
        """
        A Ponos task failure should update `Started` worker activities to `Error`.
        """
        get_logs_mock.return_value = None

        self.process_2.activity_state = ActivityState.Ready
        self.process_2.save()
        self.process_2.run()
        task = self.process_2.tasks.first()

        # Create one activity per WorkerActivityState on random elements
        element1, element2, element3, element4 = self.corpus.elements.all()[:4]
        queued_activity = self.process_2.activities.create(
            worker_version=self.version_1,
            element=element1,
            state=WorkerActivityState.Queued,
        )
        processed_activity = self.process_2.activities.create(
            worker_version=self.version_1,
            element=element3,
            state=WorkerActivityState.Processed,
        )
        error_activity = self.process_2.activities.create(
            worker_version=self.version_1,
            element=element4,
            state=WorkerActivityState.Error,
        )

        # Some state changes are only allowed when a task is in a specific state,
        # so our cases list has an initial task state, a state to update the task to,
        # and an expected activity state for `started_activity`
        cases = [
            (State.Pending, State.Running, WorkerActivityState.Started),
            (State.Running, State.Error, WorkerActivityState.Error),
            (State.Running, State.Failed, WorkerActivityState.Error),
            (State.Running, State.Completed, WorkerActivityState.Started),
            (State.Stopping, State.Stopped, WorkerActivityState.Error),
        ]

        started_activity = None
        for old_task_state, new_task_state, expected_activity_state in cases:
            # Reset activity state between each case
            if started_activity:
                started_activity.delete()
            started_activity = self.process_2.activities.create(
                worker_version=self.version_1,
                element=element2,
                state=WorkerActivityState.Started,
                started=datetime.now(timezone.utc),
            )
            task.state = old_task_state
            task.agent = self.agent
            task.save()
            with self.subTest(old_state=old_task_state, new_state=new_task_state):
                resp = self.client.patch(
                    reverse("api:task-details", kwargs={"pk": task.id}),
                    data={"state": new_task_state.value},
                    HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)
                task.refresh_from_db()
                self.assertEqual(task.state, new_task_state)

                queued_activity.refresh_from_db()
                started_activity.refresh_from_db()
                processed_activity.refresh_from_db()
                error_activity.refresh_from_db()

                self.assertEqual(queued_activity.state, WorkerActivityState.Queued)
                self.assertEqual(processed_activity.state, WorkerActivityState.Processed)
                self.assertEqual(error_activity.state, WorkerActivityState.Error)
                self.assertEqual(started_activity.state, expected_activity_state)

    @patch("arkindex.ponos.serializers.TaskSerializer.get_logs")
    @patch("arkindex.ponos.tasks.notify_process_completion.delay")
    def test_task_failure_filters_activities_worker_run(self, notify_mock, get_logs_mock):
        """
        A Ponos task failure only updates worker activity with the same worker version,
        model version and worker configuration.
        """
        get_logs_mock.return_value = None
        model = Model.objects.create(name="Generic model", public=False)
        model_version = model.versions.create(hash="b" * 32, archive_hash="a" * 32, size=8)
        worker_configuration = self.worker_1.configurations.create(name="conf")
        self.run_1.configuration = worker_configuration
        self.run_1.save()

        self.process_2.activity_state = ActivityState.Ready
        self.process_2.save()
        self.process_2.run()
        task = self.process_2.tasks.first()
        task.state = State.Running
        task.agent = self.agent
        task.save()

        # Create one activity per WorkerActivityState on random elements
        element = self.corpus.elements.first()
        activity_1 = self.process_2.activities.create(
            worker_version=self.version_1,
            configuration=worker_configuration,
            model_version=None,
            element=element,
            started=datetime.now(),
        )
        activity_2 = self.process_2.activities.create(
            worker_version=self.version_1,
            configuration=None,
            model_version=model_version,
            element=element,
            started=datetime.now(),
        )
        activity_3 = self.process_2.activities.create(
            worker_version=self.version_1,
            configuration=worker_configuration,
            model_version=model_version,
            element=element,
            started=datetime.now(),
        )

        cases = [
            (None, None, None, [activity_1, activity_2, activity_3]),
            (self.run_1, model_version, None, [activity_2]),
            (self.run_1, None, worker_configuration, [activity_1]),
            (self.run_1, model_version, worker_configuration, [activity_3]),
        ]
        for worker_run, model_version, worker_configuration, expected in cases:
            self.process_2.activities.update(state=WorkerActivityState.Started)
            if worker_run:
                worker_run.model_version = model_version
                worker_run.configuration = worker_configuration
                worker_run.save()
                task.worker_run = worker_run
                task.save()

            with self.subTest(
                worker_run=worker_run,
                model_version=model_version,
                worker_configuration=worker_configuration
            ):
                resp = self.client.patch(
                    reverse("api:task-details", kwargs={"pk": task.id}),
                    data={"state": State.Error.value},
                    HTTP_AUTHORIZATION=f"Bearer {self.agent.token.access_token}",
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)
                for activity in self.process_2.activities.filter(id__in=[e.id for e in expected]):
                    self.assertEqual(activity.state, WorkerActivityState.Error)
                for activity in self.process_2.activities.exclude(id__in=[e.id for e in expected]):
                    self.assertEqual(activity.state, WorkerActivityState.Started)
